﻿-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

-- *****************************************************************************
-- BeneCast Event Handlers
-- *****************************************************************************

function BeneCast:VARIABLES_LOADED()

	self:Debug('VARIABLES_LOADED');
	local addonname, enabled, addon;
	for i = 1, GetNumAddOns() do
		addonname, _, _, enabled = GetAddOnInfo(i);
		if enabled then
			addon = self.SnapTo[addonname];
			if addon then
				self:Debug('Found '..addonname..' for snapping to');
				wv.SnapToFound[addonname] = addon;
			end
		end
	end

	if wv.Setup.snapto == nil then
		wv.Setup.snapto = 'BeneCast';
	end
	local addon = wv.SnapToFound[wv.Setup.snapto];
	if addon == nil then
		self:Print(L['Addon %1 not loaded, defaulting to %2'](wv.Setup.snapto, self.SnapTo.BeneCast.name));
		wv.Setup.snapto = 'BeneCast';
	else
		self:Print(L['Attaching to %1'](addon.name));
	end
	self.SnapTo:ApplyConfigTable();
	
	-- Wait with the snapping till everything is loaded, give it some time
	wv.AttachFrames = true;
	
end

function BeneCast:SPELLS_CHANGED(event, arg1)

	self:Debug('SPELLS_CHANGED, arg1 = '..tostring(hp._isnil(arg1,'')));

	if arg1 ~= nil then
		-- We only want spell changes after entering the world
		return;
	end
	
	-- Reload the spell data
	wv.LoadKnownSpells = true;
	self:ResetData();
	self:UpdateAllPanels();

end

function BeneCast:UNIT_AURA(event, arg1)

	self:Debug('UNIT_AURA, arg1 = '..tostring(hp._isnil(arg1,'')));

	-- Don't bother with UNIT_AURA if the player can attack the unit
	if UnitCanAttack('player',arg1) then
		return;
	end
	if wv.AttachFrames then
		self.SnapTo:SnapPanels();
	end
	self:UpdatePanelsForUnit(arg1);

end

function BeneCast:UNIT_AURASTATE(event, arg1)

	self:Debug('UNIT_AURASTATE, arg1 = '..tostring(hp._isnil(arg1,'')));
	
	-- Don't bother with UNIT_AURASTATE if the player can attack the unit
	if UnitCanAttack('player',arg1) then
		return;
	end
	if wv.AttachFrames then
		self.SnapTo:SnapPanels();
	end
	self:UpdatePanelsForUnit(arg1);

end

function BeneCast:PLAYER_AURAS_CHANGED()

	self:Debug('PLAYER_AURAS_CHANGED');
	
	if wv.AttachFrames then
		self.SnapTo:SnapPanels();
	end
	self:UpdatePanelsForUnit('player');

end

function BeneCast:PARTY_MEMBERS_CHANGED()

	self:Debug('PARTY_MEMBERS_CHANGED');

	-- Attach the panels again, just to be sure
	self.SnapTo:SnapPanels();
	
	-- Redraw the buttons for all partymembers (and partypets)
	for i = 1,4 do
		self:AddUnitToUpdate('party' .. i, GetTime() + 1);
		self:AddUnitToUpdate('partypet' .. i, GetTime() + 1);
	end

end

function BeneCast:UNIT_PET(event, arg1)

	self:Debug('UNIT_PET, arg1 = '..tostring(hp._isnil(arg1,'')));
	
	-- Attach the panels again, just to be sure
	self.SnapTo:SnapPanels();
	
	local pet = const.PetsByOwner[arg1];
	if pet then
		self:AddUnitToUpdate(pet,GetTime() + 1);
	end

end

function BeneCast:PLAYER_TARGET_CHANGED()

	self:Debug('PLAYER_TARGET_CHANGED');
	
	-- Attach the panels again, just to be sure
	self.SnapTo:SnapPanels();
	
	-- Redraw the buttons for the target
	self:UpdatePanelsForUnit('target');
	self:AddUnitToUpdate('target',GetTime() + 1);
	
	-- Redraw the buttons for the target of target
	self:UpdatePanelsForUnit('targettarget');
	self:AddUnitToUpdate('targettarget',GetTime() + 1);

end

function BeneCast:PLAYER_FOCUS_CHANGED()
	
	self:Debug('PLAYER_FOCUS_CHANGED');
	
	-- Redraw the buttons for the focus
	self:UpdatePanelsForUnit('focus');
	self:AddUnitToUpdate('focus',GetTime() + 1);
	
end

function BeneCast:UNIT_TARGET(event, arg1)

	self:Debug('UNIT_TARGET, arg1 = '..tostring(hp._isnil(arg1,'')));
	
	if arg1 ~= 'target' then
		return;
	end
	
	-- Redraw the buttons for the target of target
	self:UpdatePanelsForUnit('targettarget');
	self:AddUnitToUpdate('targettarget',GetTime() + 1);
	
end

function BeneCast:PLAYER_ALIVE()
	BeneCast:UpdateAllPanels();
end

BeneCast.PLAYER_DEAD = BeneCast.PLAYER_ALIVE;
BeneCast.PLAYER_UNGHOST = BeneCast.PLAYER_ALIVE;

function BeneCast:UPDATE_SHAPESHIFT_FORMS()

	if wv.EffectInfo == nil then
		self:Debug('UPDATE_SHAPESHIFT_FORMS: Effects-table not loaded yet');
		return;
	else
		self:Debug('UPDATE_SHAPESHIFT_FORMS');
	end
	local stance = 'none';
	local name, active, spellinfo;
	for i = 1, GetNumShapeshiftForms() do
		_, name, active = GetShapeshiftFormInfo(i);
		if active then
			effectinfo = wv.EffectInfo[name];
			if effectinfo then
				if effectinfo.type == 'stance' then
					stance = effectinfo.group;
				end
			end
			do break end;
		end
	end
	wv.CurrentStance = stance;
	self:UpdateAllPanels();

end

function BeneCast:UNIT_HEALTH(event, arg1)

	self:Debug('UNIT_HEALTH, arg1 = '..tostring(hp._isnil(arg1,'')));
	
	-- Redraw a member's panel if they're dead
	if ( UnitHealth(arg1) <= 0 and UnitIsConnected(arg1) ) then
		self:UpdatePanelsForUnit(arg1);
	end

end

function BeneCast:PLAYER_ENTERING_WORLD()

	self:Debug('PLAYER_ENTERING_WORLD');

	-- For some unknown reason, when doing reloadui, it will not parse the OnLoad from
	-- 	the XML file for BeneCast_Tooltip... so we're doing that here.
	if not BeneCast_Tooltip:IsOwned(WorldFrame) then
		BeneCast_Tooltip:SetOwner(WorldFrame, 'ANCHOR_NONE');
	end

	-- Attach the panels again, just to be sure
	self.SnapTo:SnapPanels();
	
	self:RegisterEvent('SPELLS_CHANGED');
	self:RegisterEvent('PARTY_MEMBERS_CHANGED');
	self:RegisterEvent('PLAYER_TARGET_CHANGED');
	self:RegisterEvent('UNIT_AURA');
	self:RegisterEvent('PLAYER_AURAS_CHANGED');
	self:RegisterEvent('UNIT_AURASTATE');
	self:RegisterEvent('UPDATE_SHAPESHIFT_FORMS');
	self:RegisterEvent('PLAYER_ALIVE');
	self:RegisterEvent('PLAYER_DEAD');
	self:RegisterEvent('PLAYER_UNGHOST');
	self:RegisterEvent('UNIT_HEALTH');
	self:RegisterEvent('UNIT_PET');
	
	BeneCast.Timers.CheckEnchants:Enable();
	
	-- Reload the spell data
	wv.LoadKnownSpells = true;
	self:ResetData();
	self:UpdateAllPanels();

end

function BeneCast:PLAYER_LEAVING_WORLD()

	self:Debug('PLAYER_LEAVING_WORLD');

	self:UnregisterEvent('SPELLS_CHANGED');
	self:UnregisterEvent('PARTY_MEMBERS_CHANGED');
	self:UnregisterEvent('PLAYER_TARGET_CHANGED');
	self:UnregisterEvent('UNIT_AURA');
	self:UnregisterEvent('PLAYER_AURAS_CHANGED');
	self:UnregisterEvent('UNIT_AURASTATE');
	self:UnregisterEvent('UPDATE_SHAPESHIFT_FORMS');
	self:UnregisterEvent('PLAYER_ALIVE');
	self:UnregisterEvent('PLAYER_DEAD');
	self:UnregisterEvent('PLAYER_UNGHOST');
	self:UnregisterEvent('UNIT_HEALTH');
	self:UnregisterEvent('UNIT_PET');
	
	BeneCast.Timers.CheckEnchants:Disable();
	
end

function BeneCast:PLAYER_REGEN_DISABLED()

	self:Debug('PLAYER_REGEN_DISABLED');

	-- Player entering combat
	wv.ConsiderInCombat = true;
	self:UpdateAllPanels();
	
end

function BeneCast:PLAYER_REGEN_ENABLED()

	self:Debug('PLAYER_REGEN_ENABLED');

	-- Player exiting combat
	wv.ConsiderInCombat = false;
	self:UpdateAllPanels();
	
end
