-- Handy shortcuts
local wv = BeneCast.WorkVars;
local BeneCastButton = BeneCast.Button;
local BeneCastPanel = BeneCast.Panel;
local BeneCastPanelDropDown = BeneCast.Panel.DropDown;
local hp = BeneCast.Helper;
local L = BeneCast.Localisation;
local const = BeneCast.Constants;

-- Function to set orientation of buttons
function BeneCastPanel:OrientButtons()
	
	if self.unit == nil then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): unit is nil');
		return;
	end

	if self.pf == nil then
		BeneCast:Debug('BeneCast.Panel:OrientButtons(): pf is nil, unit is '..self.unit);
		return;
	end
	
	local profile, panelconf
	profile, panelconf = BeneCast:GetProfile(self.pf, self.unit);
	if profile == nil or panelconf == nil then
		return;
	end
	
	if not profile.orientation then
		profile.orientation = 'RIGHT';
	end
	
	if not panelconf.orientation then
		panelconf.orientation = profile.orientation;
	end

	if InCombatLockdown() then
		return;
	end

	local previousbutton = self;
	for i, button in ipairs(self.buttons) do
		
		button:ClearAllPoints();
		if panelconf.orientation == 'RIGHT' then
			button:SetPoint('LEFT',previousbutton,'RIGHT',3,0);
		elseif panelconf.orientation == 'DOWN' then
			button:SetPoint('TOP',previousbutton,'BOTTOM',0,-3);
		elseif panelconf.orientation == 'UP' then
			button:SetPoint('BOTTOM',previousbutton,'TOP',0,3);
		elseif panelconf.orientation == 'LEFT' then
			button:SetPoint('RIGHT',previousbutton,'LEFT',-3,0);
		end
		
		previousbutton = button;
				
	end
	
	self:ResizeButtons();

end

-- Function to resize the buttons of one panel
function BeneCastPanel:ResizeButtons()

	if InCombatLockdown() then
		return;
	end
	
	if self.unit == nil then
		BeneCast:Debug('BeneCast.Panel:ResizeButtons(): unit is nil');
		return;
	end
	
	if self.pf == nil then
		BeneCast:Debug('BeneCast.Panel:ResizeButtons(): pf is nil, unit is '..self.unit);
		return;
	end
	
	local profile = BeneCast:GetProfile(self.pf, self.unit);
	if profile == nil then
		return;
	end
	
	local buttonsize = profile.buttonsize;
	local button = self.next;

	-- Resize the buttons based on the saved size for every button in a panel
	for i, button in pairs(self.buttons) do
	  button:Resize(buttonsize);
	end

end

function BeneCastPanel:OnHide()

	if ( self.isMoving ) then
		self:StopMovingOrSizing();
		self.isMoving = false;
	end
	
end

function BeneCastPanel:OnLoad()

	-- set our default colors
	self:SetBackdropBorderColor(TOOLTIP_DEFAULT_COLOR.r, TOOLTIP_DEFAULT_COLOR.g, TOOLTIP_DEFAULT_COLOR.b);
	self:SetBackdropColor(TOOLTIP_DEFAULT_BACKGROUND_COLOR.r, TOOLTIP_DEFAULT_BACKGROUND_COLOR.g, TOOLTIP_DEFAULT_BACKGROUND_COLOR.b);
	self.locked = 1;

end

function BeneCastPanel:OnMouseDown(arg1)

	local profile = BeneCast:GetProfile(self.pf);
	if ( ( ( not self.locked ) or ( self.locked == 0 ) or profile.unlockbuttons ) and ( arg1 == 'LeftButton' ) ) then
		self:StartMoving();
		self.isMoving = true;
	end
	
end

function BeneCastPanel:OnMouseUp(arg1)

	if arg1 == "RightButton" then
		ToggleDropDownMenu(1, nil, self.DropDown, self:GetName(), 0, 0);
	end
	if arg1 == "LeftButton" then
		local profile, panelpf = BeneCast:GetProfile(self.pf, self.unit);
		if ( self.isMoving ) then
			self:StopMovingOrSizing();
			self.isMoving = false;
			if panelconf then
				panelpf.point = 'TOPLEFT';
				panelpf.relativepoint = 'TOPLEFT';
				local parent = self:GetParent();
				-- The values stored are unscaled values (real pixels)
				-- When being applied they will need to be adjusted to the scale between the frame and his parent
				panelpf.x = ( self:GetLeft() * self:GetEffectiveScale() ) - 
							( parent:GetLeft() * parent:GetEffectiveScale() ) + 0.5;
				panelpf.x = math.floor(panelpf.x);
				panelpf.y = ( self:GetTop() * self:GetEffectiveScale() ) -
							( parent:GetTop() * parent:GetEffectiveScale() ) + 0.5;
				panelpf.y = math.floor(panelpf.y);
				-- Move the panel to where we saved it, this so any errors in above code or the placement code
				--  would be immediately apparent instead of on next login
				BeneCast.SnapTo:SnapPanel(parent, self, panelpf);
			else
				self:Hide();
			end
		end
	end
	
end

function BeneCastPanel:OnEnter()
	if wv.Setup.debug then
		local unitname, framename;
		
		framename = this:GetName();
		unitname = UnitName(self.unit);
		if not unitname then
			unitname = 'nil';
		end
		BeneCast:Debug('Mousing over ' .. framename .. ', unit = ' .. self.unit .. ', name is ' .. unitname);
	end
end

-- Function to create a new Panel based on the xml-template
-- Panel is stored in BeneCast.WorkVars.Panels[profile][unit]
function BeneCastPanel:NewPanel(pfid, unit)

	if pfid == nil then
		error('pfid = nil');
	end

	local panelname = 'BeneCastPanel_' .. pfid .. '_' .. unit;
	local panel = CreateFrame('Frame', panelname, UIParent, 'BeneCastPanelTemplate');
	hp._CopyTableToExisting(self,panel,true);
	panel:SetScript('OnHide',panel.OnHide);
	panel:SetScript('OnLoad',panel.OnLoad);
	panel:SetScript('OnMouseDown',panel.OnMouseDown);
	panel:SetScript('OnMouseUp',panel.OnMouseUp);
	panel:SetScript('OnEnter',panel.OnEnter);
	panel.DropDown = getglobal(panelname .. 'DropDown');
	hp._CopyTableToExisting(self.DropDown,panel.DropDown,true);
	panel.DropDown.panel = panel;
	panel.DropDown:SetScript('OnLoad',panel.DropDown.OnLoad);
	panel.DropDown:SetScript('OnShow',panel.DropDown.OnShow);
	panel.pf = pfid;
	panel.unit = unit;
	panel:SetAttribute('unit',unit);
	panel:Hide();
	panel:SetFrameStrata('LOW');
	panel:SetWidth(10);
	panel:SetHeigth(10);
	if wv.Panels[pfid] == nil then
		wv.Panels[pfid] = hp.new();
	end
	wv.Panels[pfid][unit] = panel;
	return panel;

end

function BeneCastPanel:ScanButtons()
	
	wipe(self.buttons);
	local button = self.next;
	local seqno = 1;
	while button do
		tinsert(self.buttons, button);
		button.seqno = seqno;
		seqno = seqno + 1;
		button = button.next;
		if seqno > 100 then -- detect infinite loop
			error('Infinite loop detected');
		end
	end
	
end

-- Function used to update buttons, works for all possible units
function BeneCastPanel:Update()

	if self.unit == nil then
		BeneCast:Debug('BeneCastPanel:Update(): unit is nil');
		return;
	end
	
	if self.pf == nil then
		BeneCast:Debug('Panel for unit ' .. unit .. ' has no profile');
		if self:IsShown() and not InCombatLockDown() then
			self:Hide();
		end
		return;
	end
	
	if self.pf ~= BeneCast:GetCurrentProfileID() then
		if self:IsShown() and not InCombatLockdown() then
			self:Hide();
		end
		return;
	end
	
	local profile = BeneCast:GetProfile(self.pf);
	
	if not wv.AttachedTargets[self.unit] then
		if unit == 'player' or unit == 'target' or unit == 'targettarget' then
			BeneCast:Debug('Not updating '..self.unit..'-panel because it was not previously attached.');
		end
		return;
	end
	wv.AttachedTargets[self.unit] = false;
	
	if not InCombatLockdown() then
		self:Show();
	end
	
	-- Unit statusses
	local disableall = false;
	local unitexists = true;
	local unitisfriend = true;
	local hidepanel = false;

	if ( unit == 'targettarget' and not profile.showtargettarget ) then
		disableall = true;
		hidepanel = true;
	end
	
	-- if this member is non existant the parent frame will hide the panel for us
	if not UnitExists(self.unit) then
		unitexists = false;
		disableall = true;
	end
		
	local name = UnitName(self.unit);
	
	-- Hide this BeneCastPanel frame and return if this member is a pet and ShowPets is not enabled
	-- Except if the player is directly targetting the pet
	if ( self.unit ~= 'target' and self.unit ~= 'targettarget' ) and unitexists and hp._UnitIsPetID(self.unit) then
		if not profile.showpets then
			disableall = true;
			hidepanel = true;
		end
	end
	
	-- This is a hide due to settings, hide and exit if possible
	if hidepanel then
		if not InCombatLockdown() then
			self:Hide();
			return;
		end
	end
	
	-- Remember if member is mindcontrolled
	local curemagic = false;
	if unitexists then
		if UnitIsCharmed(self.unit) then
			curemagic = true;
		end
	end
	
	-- Disable this BeneCastPanel frame if member is not friendly
	if unitexists and (self.unit == 'target' or self.unit == 'targettarget') then
		if ( not UnitIsFriend('player', self.unit) and not curemagic ) then
			disableall = true;
			unitisfriend = false;
		end
	end
	
	-- If the target is invalid due to a temporary game-status keep it hidden if not in combat
	if not unitisfriend or not unitexists or disableall then
		if not InCombatLockdown() and not wv.ConsiderInCombat then	
			self:Hide();
			return;
		end
	end
	
	if unitexists then
		if ( UnitIsGhost('player') or not UnitIsConnected(self.unit) ) then
			disableall = true;
		end
	end
	
	-- If the player can ressurect this member then show the res button and return
	-- 	If there is a buff showing (UnitBuff()) then the player _cannot_ be dead, or they
	-- 	are a priest in Spirit of Redemption form, and so they shouldn't be ressurected.
	-- 	DEAD = No buffs.  Hunters do not need a special case here.
	-- 	We don't bother with ghosts, because we can't cast it on their name, we have to find
	-- 	their body.
	local showres = false;
	if unitexists and unitisfriend then
		if ( UnitIsDead(self.unit) and not UnitBuff(self.unit, 1) ) then
			showres = true;
		end
	end
	
	-- Otherwise show the buttons. Show the BeneCastPanel frame first
	if not InCombatLockdown() then
		self:Show();
	end
	
	local effectsinfo;
	local overrides = hp.new();
	local enables = hp.new();
	
	-- Check to see if the unit can be cured of a harmful effect
	-- Iterator for debuffs
	local i = 1;
	local t;
	
	if unitexists and unitisfriend then
		local debuffname, dispeltype, duration;	
		debuffname, _, _, _, dispeltype, duration = UnitDebuff(self.unit, i);
		while debuffname do
			
			local effectinfo = wv.EffectInfo[debuffname];
			if effectinfo then
				if effectinfo.overrides then
					for k, name in pairs(effectinfo.overrides) do
						t = hp.new();
							t.duration = duration;
						overrides[name] = t;
					end
				end
				if effectinfo.enables then
					for k, name in pairs(effectinfo.enables) do
						t = hp.new();
							t.duration = duration;
						enables[name] = t;
					end
				end
			end
			if dispeltype == 'Poison' then
				for k, bname in pairs(wv.DispelInfo.poison) do
					t = hp.new();
						t.duration = duration;
					enables[bname] = t;
				end
			elseif ( dispeltype == 'Disease' ) then
				for k, bname in pairs(wv.DispelInfo.disease) do
					t = hp.new();
						t.duration = duration;
					enables[bname] = t;
				end
			elseif ( dispeltype == 'Magic' ) then
				for k, bname in pairs(wv.DispelInfo.magic) do
					t = hp.new();
						t.duration = duration;
					enables[bname] = t;
				end
			elseif ( dispeltype == 'Curse' ) then
				for k, bname in pairs(wv.DispelInfo.curse) do
					t = hp.new();
						t.duration = duration;
					enables[bname] = t;
				end
			end
			i = i + 1;
			if i > 200 then -- detect infinite loop
				error('Infinite loop detected');
			end
			debuffname, _, _, _, dispeltype, duration = UnitDebuff(self.unit, i);
		end
	end
	
	-- If a target is charmed, enable the curemagic spells
	if curemagic then
		for k, bname in pairs(wv.DispelInfo.magic) do
			t = hp.new();
				t.duration = duration;
			enables[bname] = t;
		end
	end
	
	-- Check to see if the unit lacks castable buffs
	
	-- Reset the iterator for buffs
	i = 1;
	if unitexists and unitisfriend then
	
		-- Cycle through the Units Buffs to find if the effect is on the unit
		local buffname, duration, isMine;
		buffname, _, _, _, _, duration, _, isMine = UnitBuff(self.unit, i);
		while buffname do
		
			local effectinfo = wv.EffectInfo[buffname];
			if effectinfo then
				if effectinfo.type ~= 'stance' then -- The unit's stance doesn't matter
					if effectinfo.overrides then
						for k, name in pairs(effectinfo.overrides) do
							t = hp.new();
								t.duration = duration;
							overrides[name] = t;
						end
					end
					if effectinfo.enables then
						for k, name in pairs(effectinfo.enables) do
							t = hp.new();
								t.duration = duration;
							enables[name] = t;
						end
					end
				end
				if effectinfo.uninteractible and self.unit ~= 'player' then
					-- unit is uninteractible, disable all buttons
					disableall = true;
				end
			end
			
			i = i + 1;
			if i > 200 then -- detect infinite loop
				error('Infinite loop detected');
			end
			buffname, _, _, _, _, duration, _, isMine = UnitBuff(self.unit, i);
		end
	end
	
	-- Get weapon enchants
	if ( self.unit == 'player') then
		if MHWeaponEnchant ~= nil then
			t = hp.new();
				t.duration = MHWeaponEnchant.duration;
			overrides[MHWeaponEnchant.group] = t;
		end
		if OHWeaponEnchant ~= nil then
			t = hp.new();
				t.duration = OHWeaponEnchant.duration;
			overrides[OHWeaponEnchant.group] = t;
		end
	end
	
	local showallbuttons;
	if InCombatLockdown() or wv.ConsiderInCombat then
		showallbuttons = true
	else
		showallbuttons = profile.showallbuttons;
	end
	
	-- Get the config per unit type
	local config, configname;
	if self.unit == 'player' then
		configname = 'player';
	elseif self.unit == 'pet' then
		configname = 'pet';
	elseif hp._UnitIsPartyPet(self.unit) then
		configname = 'partypet';
	elseif UnitPlayerOrPetInParty(self.unit) then
		configname = 'party';
	else
		configname = 'other';
	end
	config = profile.config[configname];
	
	-- Build the buttons from the config
	local buttons = hp.new();
	
	local stanceinfo = wv.EffectInfo[wv.CurrentStance];
	
	for buttonname, setting in pairs(config) do
		if setting.enabled then
			local action, show, flash, duration, overridden, enabled;
			action = hp._CopyTable(wv.ButtonInfo[buttonname],true);
			
			-- Double-check if the action can be executed on the target
			if action and action.target and const.TargetTypes[action.target][configname] then
				-- the code first assumes that a button will be completely hidden (not faded) if it's overriden
				-- then at the end, if showallbuttons is true, a button that has .show = false is faded instead
				overridden = overrides[buttonname];
				-- if the action is not allowed in the current stance it is marked as 'overridden'
				if not overridden and stanceinfo then
					if hp._GetKeyForVal(stanceinfo.enables, buttonname) == nil then
						overridden = hp.new();
					end
				end
				enabled = enables[buttonname];
				if overridden then
					duration = overridden.duration;
				elseif enabled then
					duration = enabled.duration;
				end
				flash = (overridden or enabled) and action.flash;
				if action.type == 'rez' then
					if not wv.ConsiderInCombat and showrez then
						show = true;
					end
				elseif action.type == 'brez' then
					if wv.ConsiderInCombat and showrez then
						show = true;
					end
				elseif action.type == 'cure' or action.type == 'consume' then
					-- these need to be specifically enabled
					show = enabled and not overridden;
				else
					show = not overridden; -- overrides have priority over enables
				end
				if show or showallbuttons then
					if profile.flashasfade then
						action.show = (show and not flash);
						action.flash = nil;
					else
						action.show = show;
						action.flash = flash;
					end
					action.duration = duration;
					action.setting = hp._CopyTable(setting);
					table.insert(buttons,action);
				end
			end
		end
	end
	
	-- Sort the buttons based on their type, sorting property and name
	table.sort(buttons,function(a,b)
		local atypekey, btypekey;
		if a.type then
			atypekey = hp._GetKeyForVal(const.TypesSorted,a.type);
		end
		if b.type then
			btypekey = hp._GetKeyForVal(const.TypesSorted,b.type);
		end
		-- Odd ones go in the back
		if atypekey == nil then
			atypekey = 1000;
		end
		if btypekey == nil then
			btypekey = 1000;
		end
		if atypekey < btypekey then
			return true;
		elseif atypekey == btypekey then
			if a.sorting < b.sorting then
				return true;
			elseif a.sorting == b.sorting then
				return (a.name < b.name);
			end
		end
		return false;
	end);
	
	-- Change the panel

	-- Clear the buttons that flash for this panel
	if not wv.ButtonsThatFlash[self.pf] then
		wv.ButtonsThatFlash[self.pf] = hp.new();
	end
	wv.ButtonsThatFlash[self.pf][self.unit] = hp._ClearTable(wv.ButtonsThatFlash[self.pf][self.unit], false);
	
	local button, prevbutton;
	prevbutton = self;
	for i, action in pairs(buttons) do
		if prevbutton == nil then
			error('prevbutton = nil, i = '..tostring(i)..', unit = '..self.unit);
		end
		button = prevbutton.next;
		if not InCombatLockdown() then
			if not button then
				-- This either makes a brand new button or gets one back from retirement
				button = BeneCastButton:NewButton(self,prevbutton);
				button:Show();
			end
			button:SetAction(action);
		end
		if button then
			button:SetFade(action.show);
			if action.flash then
				BeneCast:AddButtonThatFlashes(button);
			end
		end
		prevbutton = button;
	end
	-- The rest of the buttons are retired, or faded when in combat-lockdown
	button = prevbutton.next;
	i = 1;
	while button do
		nextbutton = button.next;
		if InCombatLockdown() then
			button:SetFade(false);
		else
			button:Retire();
		end
		button = nextbutton;
		i = i + 1;
		if i > 100 then -- detect infinite loop
			error('Infinite loop detected');
		end
	end
	
	-- There is no reason that we would hide this later, so we can check this panel each
	-- 	time we have a reason to suspect a change.
	wv.AttachedTargets[self.unit] = true;
	
	-- Resize the BeneCastPanelFrame based on the number of buttons
	self:OrientButtons();
	
	hp.del(overrides);
	hp.del(enables);
	hp.del(buttons);
		
end

function BeneCastPanel:SetLock() 

	self.locked = not self.locked;
	
end

-- Function to rotate a panel clockwise
function BeneCastPanel:SetOrientation()

	BeneCast:Debug('BeneCastPanel:SetOrientation() called');
	local pf, panelpf = BeneCast:GetProfile(self.pf, self.unit);
	if pf == nil or panelpf == nil then
		self:Hide();
	else
		if panelpf.orientation == nil then
			panelpf.orientation = 'RIGHT';
		end
		-- The table contains the next value in the rotation
		panelpf.orientation = const.Orientations[panelpf.orientation].next;
		
		BeneCast:Debug('New value wv.Profiles[' .. self.pf .. '][' .. self.unit .. '].orientation = ' ..
			panelpf.orientation);
		self:OrientButtons();
	end

end

-- *****************************************************************************
-- BeneCastPanelMenu functions
-- *****************************************************************************

function BeneCastPanelDropDown:Initialize()
	
	local pf, panelpf = BeneCast:GetProfile(self.panel.pf, self.panel.unit);
	if pf == nil or panelpf == nil then
		return;
	end
	local orientation = panelpf.orientation;
	if orientation == nil then
		orientation = L['Undefined'];
	else
		orientation = const.Orientations[orientation].translation;
	end
	local info = hp.new();
	
	-- Add Title Button
	if ( UnitExists(self.panel.unit) ) then
		info.text = UnitName(self.panel.unit);
	else
		info.text = L['Unknown'];
	end
	info.isTitle = 1;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	-- Add Lock Button
	hp._ClearTable(info, false);
	info.text = L['Locked'];
	info.value = self.panel;
	info.checked = self.panel.locked;
	info.func = self.panel.SetLock;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	-- Add Orientation Button
	hp._ClearTable(info, false);
	info.text = L['Change Orientation (%1)'](orientation);
	info.value = self.panel;
	info.func = self.panel.SetOrientation;
	UIDropDownMenu_AddButton(info, UIDROPDOWNMENU_MENU_LEVEL);
	
	hp.del(info);

end

function BeneCastPanelDropDown:OnLoad()

	UIDropDownMenu_Initialize(self, self.Initialize, 'MENU');

end

function BeneCastPanelDropDown:OnShow()

	UIDropDownMenu_Initialize(self, self.Initialize, 'MENU');

end